package AccountManager::WebRequest;

use strict;
use warnings;

use CGI;
use DateTime;
use English qw(-no_match_vars);
use Template;
use Log::Any::Adapter;
use List::MoreUtils qw(uniq);

use AccountManager::Account;
use AccountManager::Account::Manager;
use AccountManager::Metadata;
use AccountManager::Service;
use AccountManager::Token;
use AccountManager::Tools;

## Defining parameters format
my $urn_or_url_regex = '(http(s?):\/\/|urn:)[^\\\$\*\"\'\`\^\|\<\>\n\s]+'
  ;    ## Format de type URL HTTP ou URN
my $url_regex     = 'http(s?):\/\/[^\\\$\*\"\'\`\^\|\<\>\n\s]+';
my $email_regex   = '([\w\-\_\.\/\+\=\'\&]+|\".*\")\@[\w\-]+(\.[\w\-]+)+';
my $domains_regex = '[\w\.\-]+(,[\w\.\-]+)*';
my %format        = (
    ## URL
    #'attributeauthority' => $url_regex,
    'entityid' => $urn_or_url_regex,
);

my %actions = (
    home               => 'req_home',
    select_sp          => 'req_select_sp',
    select_email       => 'req_select_email',
    complete_challenge => 'req_complete_challenge',
    create_accounts    => 'req_create_accounts',
);

## New web request
sub new {
    my ($pkg, %args) = @_;

    my $self = {
        configuration => $args{configuration},
    };

    Log::Any::Adapter->set(
        'File',
        $self->{configuration}->{log}->{file},
        log_level => $self->{configuration}->{log}->{level}
    );

    $self->{logger} = Log::Any->get_logger();

    AccountManager::DB->register_db(
        driver          => $self->{configuration}->{database}->{type},
        database        => $self->{configuration}->{database}->{name},
        host            => $self->{configuration}->{database}->{host},
        password        => $self->{configuration}->{database}->{password},
        username        => $self->{configuration}->{database}->{user}
    );

    $self->{db} = AccountManager::DB->new();
    $self->{cgi} = CGI->new();

    bless $self, $pkg;

    return $self;
}

sub run {
    my ($self) = @_;

    # initialize output parameters
    $self->{out} = {
        env => {
            REMOTE_HOST => $ENV{REMOTE_HOST},
            REMOTE_ADDR => $ENV{REMOTE_ADDR},
            SCRIPT_NAME => $ENV{SCRIPT_NAME}
        },
        conf => {
            accounts_validity_period => $self->{configuration}->{service}->{account_validity_period},
            app_name                 => $self->{configuration}->{app}->{name},
            app_url                  => $self->{configuration}->{app}->{url},
            idp_scope                => $self->{configuration}->{idp}->{scope},
            idp_displayname          => $self->{configuration}->{idp}->{displayname},
            support_email            => $self->{configuration}->{app}->{support_email},
            version                  => $self->{configuration}->{app}->{version},
        },
        title  => $self->{configuration}->{app}->{name}
    };

    # process input parameters
    my %parameters = $self->{cgi}->Vars();

    foreach my $parameter (keys %parameters) {

        # cleanup
        $parameters{$parameter} =~ s/\r//g;  # remove &0D char
        $parameters{$parameter} =~ s/\s+$//; # remove trailing spaces
        $parameters{$parameter} =~ s/^\s+//; # remove leading spaces

        # format check
        if (defined $format{$parameter}
            && !ref($format{$parameter})) {
            if ($parameters{$parameter} !~ /^$format{$parameter}$/) {
                push @{ $self->{out}->{errors} }, "format_$parameter";
                $self->{logger}->error(
                    "Incorrect parameter format : $parameter"
                );
                $self->respond();
            }
        }

        # If action_xx parameter is set, set action parameter with value xx
        if ($parameter =~ /^action_(\w+)$/) {
            $parameters{action} = $1;
        }

        # register needed parameters
        $self->{in} = {
            email    => $parameters{email},
            style    => $parameters{style},
            entityid => $parameters{entityid},
            token    => $parameters{token}
        };
    }

    # process requested action
    my $action = $parameters{action} || 'home';
    $self->{out}->{action} = $action;
    if ($actions{$action}) {
        my $method = $actions{$action};
        $self->$method();
    } else {
        ## unknown action
        push @{ $self->{out}->{errors} }, "unknown_action";
        $self->{logger}->error( "Unknown action '$action'");
        $self->respond();
    }

    return 1;
}

## Return HTML content
sub respond {
    my ($self) = @_;

    ## Parse template
    my $tt2 = Template->new({
        INCLUDE_PATH => $self->{configuration}->{_}->{templates_dir}
    });

    my $template;

    ## nobanner is used to do AJAX to get only pieces of HTML to load in the web client
    if ($self->{in}->{style} && $self->{in}->{style} eq 'nobanner') {
        $template = 'web/index-nobanner.tt2.html';
    } else {
        $template = 'web/index.tt2.html';
    }

    binmode(STDOUT, ":utf8");

    print $self->{cgi}->header(
        -type    => 'text/html',
        -charset => ''
    );

    unless ($tt2->process($template, $self->{out}, \*STDOUT)) {
        printf "Content-type: text/plain\n\n Error: %s", $tt2->error();
        $self->{logger}->errorf("Web parser error : %s", $tt2->error());
    }

}

sub req_select_sp {
    my ($self) = @_;

    my $metadata;

    eval {
        $metadata = AccountManager::Metadata->new(
            file => $self->{configuration}->{_}->{federation_metadata_file}
        );
    };
    if ($EVAL_ERROR) {
        push @{ $self->{out}->{errors} }, "internal";
        $self->{logger}->error("Failed to load federation metadata: $EVAL_ERROR");
        $self->respond();
    }

    $self->{out}->{metadata} = $metadata->parse(type => 'sp');

    $self->respond();
}

sub req_select_email {
    my ($self) = @_;

    unless ($self->{in}->{entityid}) {
        push @{ $self->{out}->{errors} }, "missing_entityid";
        $self->{logger}->error("Missing parameter entityid");
        $self->respond();
    }

    # Create a persistent service provider object
    my $provider = AccountManager::Service->new(
        db       => $self->{db},
        entityid => $self->{in}->{entityid}
    );

    if ($provider->load(speculative => 1)) {
        # already present in DB, nothing todo
    } else {
        # extract information from metadata
        my $metadata;

        eval {
            $metadata = AccountManager::Metadata->new(
                file => $self->{configuration}->{_}->{federation_metadata_file}
            );
        };
        if ($EVAL_ERROR) {
            push @{ $self->{out}->{errors} }, "internal";
            $self->{logger}->error("Failed to load federation metadata: $EVAL_ERROR");
            $self->respond();
        }

        my $sps = $metadata->parse(id => $self->{in}->{entityid});
        if (!@$sps) {
            push @{ $self->{out}->{errors} }, "no_such_entity";
            $self->{logger}->errorf(
                "No such SP '%s' in metadata", $self->{in}->{entityid}
            );
            $self->respond();
        }
        my $sp = $sps->[0];

        # complete persistent object
        $provider->displayname($sp->{display_name});
        $provider->contacts(uniq map { $_->{EmailAddress} } @{$sp->{contacts}})
            if $sp->{contacts};

        # save in DB
        unless ($provider->save()) {
            push @{ $self->{out}->{errors} }, "internal";
            $self->{logger}->error("Failed to save service provider object");
            $self->respond();
        }
    }

    # override metadata contacts if needed
    my $entity = $self->{in}->{entityid};
    my $contacts =
        $self->{configuration}->{$entity}->{contacts} ||
        $self->{configuration}->{service}->{contacts};
    if ($contacts) {
        if ($contacts =~ /^\+(.+)/) {
            # complement original contacts
            $provider->contacts($provider->contacts(), split(/, */, $1));
        } else {
            # replace original contacts
            $provider->contacts(split(/, */, $contacts));
        }
    }

    $self->{out}->{provider} = $provider;

    $self->respond();
}

sub req_complete_challenge {
    my ($self) = @_;

    unless ($self->{in}->{entityid}) {
        push @{ $self->{out}->{errors} }, "missing_entityid";
        $self->{logger}->error("Missing parameter entityid");
        $self->respond();
    }

    unless ($self->{in}->{email}) {
        push @{ $self->{out}->{errors} }, "missing_email";
        $self->{logger}->error("Missing parameter email");
        $self->respond();
    }

    my $provider = AccountManager::Service->new(
        db       => $self->{db},
        entityid => $self->{in}->{entityid},
    );

    unless ($provider->load(speculative => 1)) {
        push @{ $self->{out}->{errors} }, "no_such_entity";
        $self->{logger}->errorf("No such SP '%s' in database", $self->{in}->{entityid});
        $self->respond();
    }

    # override metadata contacts if needed
    my $entity = $self->{in}->{entityid};
    my $contacts =
        $self->{configuration}->{$entity}->{contacts} ||
        $self->{configuration}->{service}->{contacts};
    if ($contacts) {
        if ($contacts =~ /^\+(.+)/) {
            # complement original contacts
            $provider->contacts($provider->contacts(), split(/, */, $1));
        } else {
            # replace original contacts
            $provider->contacts(split(/, */, $contacts));
        }
    }

    ## Check that email is a known contact for this SP
    unless ($provider->is_contact($self->{in}->{email}))
    {
        push @{ $self->{out}->{errors} }, "internal";
        $self->{logger}->errorf(
            "Requested a token for %s for an unautorized address '%s'",
            $self->{in}->{entityid},
            $self->{in}->{email}
        );
        $self->respond();
    }

    # delete any previous token for the same email/service couple
    my $old_token = AccountManager::Token->new(
        db            => $self->{db},
        email_address => $self->{in}->{email},
        sp_entityid   => $self->{in}->{entityid}
    );

    if ($old_token->load(speculative => 1)) {
        unless ($old_token->delete()) {
            push @{ $self->{out}->{errors} }, "internal";
            $self->{logger}->errorf(
                "Failed to delete previous authentication token with ID %s",
                $old_token->id()
            );
            $self->respond();
        }
    }

    # compute a new token
    my $validity_period =
        $self->{configuration}->{_}->{tokens_validity_period};
    my $token = AccountManager::Token->new(
        db              => $self->{db},
        email_address   => $self->{in}->{email},
        sp_entityid     => $self->{in}->{entityid},
        creation_date   => DateTime->now(),
        expiration_date => DateTime->now()->add(hours => $validity_period),
        token           => AccountManager::Tools::generate_secret(20)
    );

    unless ($token->save()) {
        push @{ $self->{out}->{errors} }, "internal";
        $self->{logger}->error("Failed to save authentication token");
        $self->respond();
    }

    $self->{out}->{email}    = $self->{in}->{email};
    $self->{out}->{entityid} = $self->{in}->{entityid};

    my $sender    = $self->{configuration}->{_}->{notice_from};
    my $recipient = $self->{in}->{email};
    my $sendmail  = $self->{configuration}->{_}->{sendmail_path} ||
                    '/usr/sbin/sendmail';

    open(my $handle, '|-', "$sendmail -f $sender $recipient") or do {
        push @{ $self->{out}->{errors} }, "mail_notification_error";
        $self->{logger}->errorf("Unable to run sendmail executable: %s", $ERRNO);
        $self->respond();
    };

    my $tt2 = Template->new({
        INCLUDE_PATH => $self->{configuration}->{_}->{templates_dir}
    });
    my $template = 'mail/send_authentication_token.tt2.eml';
    my $data = {
        env => {
            REMOTE_HOST => $ENV{REMOTE_HOST},
            REMOTE_ADDR => $ENV{REMOTE_ADDR},
        },
        conf => {
            app_name      => $self->{configuration}->{app}->{name},
            app_url       => $self->{configuration}->{app}->{url},
            support_email => $self->{configuration}->{app}->{support_email},
        },
        from     => $sender,
        to       => $recipient,
        entityid => $self->{in}->{entityid},
        token    => $token->token(),
    };

    unless ($tt2->process($template, $data, $handle)) {
        push @{ $self->{out}->{errors} }, "mail_notification_error";
        $self->{logger}->errorf("Mail notification error: %s", $tt2->error());
        $self->respond();
    }

    close $handle;

    $self->{logger}->infof(
        "Token send to %s for entityid=%s;token=%s",
        $self->{in}->{email},
        $self->{in}->{entityid},
        $token->token(),
    );

    $self->respond();
}

sub req_create_accounts {
    my ($self) = @_;

    unless ($self->{in}->{entityid}) {
        push @{ $self->{out}->{errors} }, "missing_entityid";
        $self->{logger}->error("Missing parameter entityid");
        $self->respond();
    }

    unless ($self->{in}->{token}) {
        push @{ $self->{out}->{errors} }, "missing_token";
        $self->{logger}->error("Missing parameter token");
        $self->respond();
    }

    my $token = AccountManager::Token->new(
        db    => $self->{db},
        token => $self->{in}->{token}
    );

    if (! $token->load(speculative => 1)) {
        push @{ $self->{out}->{errors} }, "wrong_token";
        $self->{logger}->errorf(
            "Failed to validate authentication token %s for entityid %s",
            $self->{in}->{token},
            $self->{in}->{entityid}
        );
        $self->respond();
    }

    if (! $token->sp_entityid() eq $self->{in}->{entityid}) {
        push @{ $self->{out}->{errors} }, "wrong_token_for_sp";
        $self->{logger}->errorf(
            "Authentication token %s cannot be used for SP with entityid %s",
            $self->{in}->{token},
            $self->{in}->{entityid}
        );
        $self->respond();
    }

    ## delete the token
    unless ($token->delete()) {
        $self->{logger}->errorf(
            "Failed to delete authentication token %s",
            $self->{in}->{token}
        );
    }

    ## create test accounts
    my @accounts;

    my $entity = $self->{in}->{entityid};
    my $profiles =
        $self->{configuration}->{$entity}->{account_profiles} ||
        $self->{configuration}->{service}->{account_profiles};
    my $validity_period =
        $self->{configuration}->{$entity}->{account_validity_period} ||
        $self->{configuration}->{service}->{account_validity_period};

    foreach my $profile (split(/, */, $profiles)) {
        my $password = AccountManager::Tools::generate_password(10);
        my $account = AccountManager::Account->new(
            db              => $self->{db},
            profile         => $profile,
            sp_entityid     => $entity,
            scope           => $self->{configuration}->{idp}->{scope},
            password        => $password,
            password_hash   => AccountManager::Tools::sha256_hash($password),
            creation_date   => DateTime->now(),
            expiration_date => DateTime->now()->add(days => $validity_period)
        );
        next unless $account->save();
        push @accounts, $account;
    }

    unless (@accounts) {
        push @{ $self->{out}->{errors} }, "accounts_creation_failed";
        $self->{logger}->errorf(
            "Failed to create test accounts for SP with entityid %s",
            $self->{in}->{entityid}
        );
        $self->respond();
    }

    ## Update simpleSAMLphp configuration to enable test accounts
    my $accounts = AccountManager::Account::Manager->get_accounts(
        db => $self->{db}
    );

    eval {
        AccountManager::Tools::update_ssp_authsources(
            $self->{configuration}->{_}->{templates_dir},
            $self->{configuration}->{idp}->{accounts_file},
            $accounts
        );
    };
    if ($EVAL_ERROR) {
        push @{ $self->{out}->{errors} }, "accounts_creation_failed";
        $self->{logger}->errorf(
            "Failed to create simpleSAMLphp configuration file: %s",
            $EVAL_ERROR
        );
        $self->respond();
    }

    $self->{logger}->infof(
        "Token validated for entityid=%s;token=%s",
        $self->{in}->{entityid},
        $self->{in}->{token}
    );

    $self->{out}->{entityid} = $self->{in}->{entityid};
    $self->{out}->{accounts} = \@accounts;

    $self->respond();
}

## Return the homepage of the service
sub req_home {
    my ($self) = @_;

    $self->respond();
}

1;
